//---------------------------------------------------
// module  : BuadesPrior.h
// auteur  : 
// date    :
// langage : C++
// labo    :
//---------------------------------------------------
//---------------------------------------------------

#ifndef BUADESPRIOR_H
#define BUADESPRIOR_H

//-----------------------------------------------
// FICHIERS INCLUS ------------------------------
//-----------------------------------------------
#include "function.h"

//------------------------------------------------
// DEFINITIONS -----------------------------------
//------------------------------------------------
#define SIZEW           7 
#define BIGSIZE        11 
#define SIGMATEX       10.0*SIGMA
#define SIGMAGK         1.0

//------------------------------------------------
// CLASSE BuadesPrior  ----------------------------------
//------------------------------------------------
class BuadesPrior
 {
   public:

  //>Img
  int  length,width;

  //Parameters
  int   Size;       
  int   BigSize;      
  float SigmaH;
  float SigmaGK;   

  //Flag
  int FlgSilent; 

  //Allocation
  float**  MatFiltered;

  float**      SquWin; 
  float**      SquWin2;   
  float**      SquWinGaussianKernel;

  float*** MatWeightX;  
  float*** MatWeightY;  
  float*** MatWeightW; 
  float**  MatDeriv;

  //>Constructeur 
  public:
  BuadesPrior(int,int,int,int,float,float);
  ~BuadesPrior();

  //Methodes
  void  ComputePrior(float**);
  void  Denoising(float**,float**);

  float computeMMSE(float**,float**,int,int);
  float funcgauss2D(float,float,float); 
  };
//---------------------------------------------

//----------------------------------------------------------
// Constructeur de la classe BuadesPrior
//---------------------------------------------------------- 
inline
BuadesPrior::BuadesPrior(int lgth,int wdth,int sz,int bsz,float sgh,float sggk)
{
 int i,j,m,n;
 float sum;

 //Enregistrement
 length=lgth;
 width=wdth;
 Size=sz;
 BigSize=bsz;
 SigmaH=sgh;
 SigmaGK=sggk;

 //Information
 printf("\n\n   BuadesPrior>Enregistrement"); 
 printf("\n       Size=     [%d]",Size); 
 printf("\n       BigSize=  [%d]",BigSize); 
 printf("\n       SigmaH=   [%.2f] (10xsigma)",SigmaH);
 //printf("\n       SigmaGK=  [%.2f]",SigmaGK); 

 //Allocation
 MatFiltered=fmatrix_allocate_2d(length,width);

 SquWin=fmatrix_allocate_2d(Size,Size);
 SquWin2=fmatrix_allocate_2d(Size,Size);
 SquWinGaussianKernel=fmatrix_allocate_2d(Size,Size);

 MatDeriv=fmatrix_allocate_2d(length,width);
 MatWeightX=fmatrix_allocate_3d(BigSize*BigSize,length,width);
 MatWeightY=fmatrix_allocate_3d(BigSize*BigSize,length,width);
 MatWeightW=fmatrix_allocate_3d(BigSize*BigSize,length,width);

 //Initialisation
 for(i=0;i<length;i++) for(j=0;j<width;j++)
   MatFiltered[i][j]=0.0;

 //Pre-Compute
 sum=0.0;
 for(m=0;m<Size;m++) for(n=0;n<Size;n++) 
   {  SquWinGaussianKernel[m][n]=funcgauss2D(abs(m-(Size/2)),abs(n-(Size/2)),SigmaGK*SigmaGK);
      sum+=SquWinGaussianKernel[m][n]; }

 //>Normalisation
 for(m=0;m<Size;m++) for(n=0;n<Size;n++) 
   SquWinGaussianKernel[m][n]=SquWinGaussianKernel[m][n]*(Size*Size)/sum;
}

//----------------------------------------------------------
// Destructeur de la classe BuadesPrior
//----------------------------------------------------------
inline
BuadesPrior::~BuadesPrior()
 { 
  //Liberation Memoire
   if (MatFiltered) free_fmatrix_2d(MatFiltered); 

   if (SquWin)  free_fmatrix_2d(SquWin);
   if (SquWin2) free_fmatrix_2d(SquWin2);
   if (SquWinGaussianKernel) free_fmatrix_2d(SquWinGaussianKernel); 

   if (MatDeriv)    free_fmatrix_2d(MatDeriv);

   if (MatWeightX) free_fmatrix_3d(MatWeightX,BigSize*BigSize);
   if (MatWeightY) free_fmatrix_3d(MatWeightY,BigSize*BigSize);
   if (MatWeightW) free_fmatrix_3d(MatWeightW,BigSize*BigSize); 
 }

//-----------------
//-----------------
//----------------------------------------------------------   
// ComputePrior
//----------------------------------------------------------   
inline  
void BuadesPrior::ComputePrior(float** mat)
{
 int   i,j,k,l,m,n;
 int   posc,posr;
 int   row,col;
 float dist;

 float distcliq=0.0;
 float sumdistcliq=0.0;

 
  //-------Non Local Regularization Graph
  //---------------------------------------------------------------
  printf("\n\n   Calcul Graph Non-Local ...");
  for(i=0;i<length;i++) 
  {
    if (!(i%50)) printf("[%d]",i); fflush(stdout);
    for(j=0;j<width;j++)  
    {
     //>Small Window
     for(k=0;k<Size;k++) for(l=0;l<Size;l++) 
        { posr=i-(Size/2)+k;
          posc=j-(Size/2)+l;
          if (posr<0)           posr+=length;
          if (posr>(length-1))  posr-=length; 
          if (posc<0)           posc+=width;
          if (posc>(width-1))   posc-=width;
          SquWin[k][l]=mat[posr][posc]; }

     //> Search In Big Window
     sumdistcliq=0.0;
     for(k=0;k<BigSize;k++) for(l=0;l<BigSize;l++) 
        { posr=i-(BigSize/2)+k;
          posc=j-(BigSize/2)+l;

          if (posr<0)           posr+=length;
          if (posr>(length-1))  posr-=length; 
          if (posc<0)           posc+=width;
          if (posc>(width-1))   posc-=width;

           for(m=0;m<Size;m++) for(n=0;n<Size;n++) 
	      { row=posr-(Size/2)+m;
	        col=posc-(Size/2)+n;
                if (row<0)           row+=length;
                if (row>(length-1))  row-=length; 
                if (col<0)           col+=width;
                if (col>(width-1))   col-=width;
                SquWin2[m][n]=mat[row][col]; }
                 
            dist=0.0;
            for(m=0;m<Size;m++) for(n=0;n<Size;n++) 
              dist+=CARRE(SquWin2[m][n]-SquWin[m][n]);

	        distcliq=exp(-dist/CARRE(SigmaH));
                if (distcliq<0.0000001) distcliq=0.0000001;
                sumdistcliq+=distcliq;
                MatWeightX[(k*BigSize)+l][i][j]=posr;
                MatWeightY[(k*BigSize)+l][i][j]=posc;
                MatWeightW[(k*BigSize)+l][i][j]=distcliq; 
	}

     MatDeriv[i][j]=(1.0-1.0/sumdistcliq);
    }
  }//---------------------------------------------------------------- 
}

//---------------
// DENOISING
//---------------
//----------------------------------------------------------   
// Denoising Buades
//----------------------------------------------------------   
inline  
void BuadesPrior::Denoising(float** matin,float** matout)
{
 int i,j,k,l;
 int posr,posc;
 float SumWeight;
 float Weight;

 //Denoising
 //-----------------------------------------------------------------
 if (!FlgSilent) printf("\n\n   Denoising ...");
 fflush(stdout);
 for(i=0;i<length;i++) for(j=0;j<width;j++)  
       {
        SumWeight=0.0;
        Weight=0.0; 
        for(k=0;k<BigSize;k++) for(l=0;l<BigSize;l++)
	   {
            posr=(int)MatWeightX[(k*BigSize)+l][i][j];
            posc=(int)MatWeightY[(k*BigSize)+l][i][j];

	    SumWeight+=MatWeightW[(k*BigSize)+l][i][j]; 
            Weight+=MatWeightW[(k*BigSize)+l][i][j]*matin[posr][posc];
            }

        matout[i][j]=Weight/SumWeight;  
  
       }//------------------------------------------------------------
}

//----------------------------------------------------------
// MMSE              
//----------------------------------------------------------    
inline   
float BuadesPrior::computeMMSE(float** mat1,float** mat2,int lgth,int wdth)
{
 int i,j;
 float mmse;

 //Boucle 
 mmse=0.0;
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) mmse+=CARRE(mat2[i][j]-mat1[i][j]);

 mmse=mmse/(lgth*wdth);

 //retour
 return mmse;
}

//----------------------------------------------------------
// Gaussienne 2D                         
//----------------------------------------------------------
inline 
float BuadesPrior::funcgauss2D(float x,float y,float var)
{
 return(exp(-(SQUARE(x)+SQUARE(y))/(2*var))/(2.0*PI*var));
}


#endif
