//------------------------------------------------------
// module  : Restoration.cc
// auteur  : Mignotte Max
// date    : 
// version : 1.0
// langage : C++
// note    :
//------------------------------------------------------
//  

//------------------------------------------------
// FICHIERS INCLUS -------------------------------
//------------------------------------------------
#include "Restoration.h"

//------------------------------------------------
// DEFINITIONS -----------------------------------
//------------------------------------------------
#define NAME_IMG      "Image"
#define NAME_DEGRAD   "ImgDegrad"
#define NAME_REST     "ImgRest"

//------------------------------------------------
// CONSTANTES ------------------------------------
//------------------------------------------------
const int   METHOD=0;
const int   PSF=1;
const int   SIZE=9;
const float VARIANCE_NOISE=0.308;

const int ZOOM=1;
const int QUIT=0;
const int SAVE=0;

//------------------------------------------------
// VARIABLES GLOBALES ----------------------------
//------------------------------------------------
char Name_img[NCHAR];
char Name_imgdeg[NCHAR];

int   method;
int   psf;
int   size;
float variance_noise;

int   nbiter;
float lambda;
float Tau;

int    seed;

int   zoom;
int   flag_save;
int   flag_quit;

//-------------------------
//--- Windows -------------
//-------------------------
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>
#include <X11/cursorfont.h>

Display   *display;
int	  screen_num;
int 	  depth;
Window	  root;
Visual*	  visual;
GC	  gc;

/************************************************************************/
/* OPEN_DISPLAY()							*/
/************************************************************************/
int open_display()
{
  if ((display=XOpenDisplay(NULL))==NULL)
   { printf("Connection impossible\n");
     return(-1); }

  else
   { screen_num=DefaultScreen(display);
     visual=DefaultVisual(display,screen_num);
     depth=DefaultDepth(display,screen_num);
     root=RootWindow(display,screen_num);
     return 0; }
}

/************************************************************************/
/* FABRIQUE_WINDOW()							*/
/* Cette fonction cre une fenetre X et l'affiche  l'cran.	        */
/************************************************************************/
Window fabrique_window(char *nom_fen,int x,int y,int width,int height,int zoom)
{
  Window                 win;
  XSizeHints      size_hints;
  XWMHints          wm_hints;
  XClassHint     class_hints;
  XTextProperty  windowName, iconName;

  char *name=nom_fen;

  if(zoom<0) { width/=-zoom; height/=-zoom; }
  if(zoom>0) { width*=zoom;  height*=zoom;  }

  win=XCreateSimpleWindow(display,root,x,y,width,height,1,0,255);

  size_hints.flags=PPosition|PSize|PMinSize;
  size_hints.min_width=width;
  size_hints.min_height=height;

  XStringListToTextProperty(&name,1,&windowName);
  XStringListToTextProperty(&name,1,&iconName);
  wm_hints.initial_state=NormalState;
  wm_hints.input=True;
  wm_hints.flags=StateHint|InputHint;
  class_hints.res_name=nom_fen;
  class_hints.res_class=nom_fen;

  XSetWMProperties(display,win,&windowName,&iconName,
                   NULL,0,&size_hints,&wm_hints,&class_hints);

  gc=XCreateGC(display,win,0,NULL);

  XSelectInput(display,win,ExposureMask|KeyPressMask|ButtonPressMask| 
               ButtonReleaseMask|ButtonMotionMask|PointerMotionHintMask| 
               StructureNotifyMask);

  XMapWindow(display,win);
  return(win);
}

/****************************************************************************/
/* CREE_XIMAGE2()							    */
/* Cre une XImage  partir d'un tableau de float                           */
/* L'image peut subir un zoom.						    */
/****************************************************************************/
XImage* cree_Ximage2(float** mat,int z,int length,int width)
{
  int lgth,wdth,lig,col,zoom_col,zoom_lig;
  float somme;
  unsigned char	 pix;
  unsigned char* dat;
  XImage* imageX;

  /*Zoom positiv*/
  /*------------*/
  if (z>0)
  {
   lgth=length*z;
   wdth=width*z;

   dat=(unsigned char*)malloc(lgth*(wdth*4)*sizeof(unsigned char));
   if (dat==NULL)
      { printf("Impossible d'allouer de la memoire.");
        exit(-1); }

  for(lig=0;lig<lgth;lig=lig+z) for(col=0;col<wdth;col=col+z)
   { 
    pix=(unsigned char)mat[lig/z][col/z];
    for(zoom_lig=0;zoom_lig<z;zoom_lig++) for(zoom_col=0;zoom_col<z;zoom_col++)
      { 
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+0)]=pix;
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+1)]=pix;
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+2)]=pix;
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+3)]=pix; 
       }
    }
  } /*--------------------------------------------------------*/

  /*Zoom negatifv*/
  /*------------*/
  else
  {
   z=-z;
   lgth=(length/z);
   wdth=(width/z);

   dat=(unsigned char*)malloc(lgth*(wdth*4)*sizeof(unsigned char));
   if (dat==NULL)
      { printf("Impossible d'allouer de la memoire.");
        exit(-1); }

  for(lig=0;lig<(lgth*z);lig=lig+z) for(col=0;col<(wdth*z);col=col+z)
   {  
    somme=0.0;
    for(zoom_lig=0;zoom_lig<z;zoom_lig++) for(zoom_col=0;zoom_col<z;zoom_col++)
     somme+=mat[lig+zoom_lig][col+zoom_col];
           
     somme/=(z*z);    
     dat[((lig/z)*wdth*4)+((4*(col/z))+0)]=(unsigned char)somme;
     dat[((lig/z)*wdth*4)+((4*(col/z))+1)]=(unsigned char)somme;
     dat[((lig/z)*wdth*4)+((4*(col/z))+2)]=(unsigned char)somme;
     dat[((lig/z)*wdth*4)+((4*(col/z))+3)]=(unsigned char)somme; 
   }
  } /*--------------------------------------------------------*/

  imageX=XCreateImage(display,visual,depth,ZPixmap,0,(char*)dat,wdth,lgth,16,wdth*4);
  return (imageX);
}

/****************************************************************************/
/* CREE_XIMAGE4()							    */
/* Cre une XImage  partir d'un tableau d'entier                           */
/* L'image peut subir un zoom.						    */
/****************************************************************************/
XImage* cree_Ximage4(int** mat,int z,int length,int width)
{
  int lgth,wdth,lig,col,zoom_col,zoom_lig;
  float somme;
  unsigned char	 pix;
  unsigned char* dat;
  XImage* imageX;

  /*Zoom positif*/
  /*------------*/
  if (z>0)
  {
   lgth=length*z;
   wdth=width*z;

   dat=(unsigned char*)malloc(lgth*(wdth*4)*sizeof(unsigned char));
   if (dat==NULL)
      { printf("Impossible d'allouer de la memoire.");
        exit(-1); }

  for(lig=0;lig<lgth;lig=lig+z) for(col=0;col<wdth;col=col+z)
   { 
    pix=(unsigned char)mat[lig/z][col/z];
    for(zoom_lig=0;zoom_lig<z;zoom_lig++) for(zoom_col=0;zoom_col<z;zoom_col++)
      { 
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+0)]=pix;
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+1)]=pix;
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+2)]=pix;
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+3)]=pix; 
       }
    }
  } /*--------------------------------------------------------*/

  /*Zoom negatif*/
  /*------------*/
  else
  {
   z=-z;
   lgth=(length/z);
   wdth=(width/z);

   dat=(unsigned char*)malloc(lgth*(wdth*4)*sizeof(unsigned char));
   if (dat==NULL)
      { printf("Impossible d'allouer de la memoire.");
        exit(-1); }

  for(lig=0;lig<(lgth*z);lig=lig+z) for(col=0;col<(wdth*z);col=col+z)
   {  
    somme=0.0;
    for(zoom_lig=0;zoom_lig<z;zoom_lig++) for(zoom_col=0;zoom_col<z;zoom_col++)
     somme+=mat[lig+zoom_lig][col+zoom_col];
           
     somme/=(z*z);    
     dat[((lig/z)*wdth*4)+((4*(col/z))+0)]=(unsigned char)somme;
     dat[((lig/z)*wdth*4)+((4*(col/z))+1)]=(unsigned char)somme;
     dat[((lig/z)*wdth*4)+((4*(col/z))+2)]=(unsigned char)somme;
     dat[((lig/z)*wdth*4)+((4*(col/z))+3)]=(unsigned char)somme; 
   }
  } /*--------------------------------------------------------*/

  imageX=XCreateImage(display,visual,depth,ZPixmap,0,(char*)dat,wdth,lgth,16,wdth*4);
  return (imageX);
}


//---------------------------------------------------------
// Enregistre les arguments de la ligne de commande      
//---------------------------------------------------------
void read_arguments(int argc,char** argv)
{
 int i;

   //Options
  if (argc<2) 
    { printf("\n Restor [Img]"); 
      printf("\n Options are :  values by default are indicated in  ..... <(...)> ");
      printf("\n ---------------------------------------------------------------- ");
      printf("\n            -t #  .....<(%d)>",PSF);
      printf("\n            -t 0 Psf -0- (-i 1) No Psf (dirac)");
      printf("\n            -t 1 Psf -1- Uniform Blur");
      printf("\n            -t 2 Psf -2- (1/1+i^2+j^2) Blur");
      printf("\n            -t 3 Psf -3- (1-4-6-4-1)/16 Blur (horizontal & vertical)");
      printf("\n            -t 4 Psf -4- (-i 9) (1-4-10-16-19-16-10-4-1)/81 Blur (horiz & vert)");
      printf("\n            -t 5 Psf -5- (-i 12) (1+(i^2+j^2/R^2))^epsilon");
      printf("\n");
      printf("\n            -i size of the PSF  .....<(%d)>",SIZE);
      printf("\n            -v variance of the noise  .....<(%.3f)>",VARIANCE_NOISE);
      printf("\n            -------------");
      printf("\n            -m # Proposed Methods .....<(%d)>",METHOD);
      printf("\n            -m 0  Unsup N-L Rest2");
      printf("\n            --------------");
      //printf("\n            -r NbIter  (%d)",ITERATION_REST);
      //printf("\n            -l Gamma [%.2f]",LAMBDA);
      //printf("\n            -u Tau (%.2f)",(float)TAU);
      printf("\n            -------------");
      printf("\n            -z zoom (%d)",ZOOM);
      printf("\n            -q quit (without)"); 
      printf("\n            -s save (without)"); 
      printf("\n            -------------");
      printf("\n");
      printf("\n        Examples: For Exp1, Exp2, Exp3, Exp4, Exp5 and Exp6 (see paper)");
      printf("\n        --------------------------------------------------------------");
      printf("\n   Restoration cameraman.pgm -t 1 -i 9 -v 0.308"); 
      printf("\n   Restoration cameraman.pgm -t 2 -i 7 -v 2.0"); 
      printf("\n   Restoration cameraman.pgm -t 2 -i 7 -v 8.0"); 
      printf("\n   Restoration lena512.pgm -t 3 -i 5 -v 49.0"); 
      printf("\n   Restoration cameraman.pgm -t 1 -i 5 -v 33.3");
      printf("\n   Restoration cameraman.pgm -t 5 -i 9 -v 62.5");
      printf("\n\n\n\n");
      exit(-1); }

      //Chargement Fichier
      if (argc>3) strcpy(Name_img,argv[1]);
     

   for(i=2;i<argc;i++)
     {   
          switch(argv[i][1])
          {
           case 't': psf=atoi(argv[++i]);                  break;
           case 'v': variance_noise=atof(argv[++i]);       break;
           case 'i': size=atoi(argv[++i]);                 break;

           case 'm': method=atoi(argv[++i]);               break;

           case 'r': nbiter=atoi(argv[++i]);               break;
           case 'l': lambda=atof(argv[++i]);               break;
           case 'u': Tau=atof(argv[++i]);                  break; 
 
           case 'e': seed=atoi(argv[++i]);                 break;
	             
           case 'z': zoom=atoi(argv[++i]);                 break;
	   case 'q': flag_quit=1;                          break;
           case 's': flag_save=1;                          break;                   
          }    
     } 
 }

//----------------------------------------------------------
//----------------------------------------------------------
// PROGRAMME PRINCIPAL -------------------------------------
//----------------------------------------------------------
//----------------------------------------------------------
int main(int argc,char** argv)
{
 int   flag;

 //Pour Xwindow
 //------------
 XEvent ev;
 Window win_ppicture_img;
 Window win_ppicture_deg;
 Window win_ppicture_rest;
 
 XImage *x_ppicture_img;
 XImage *x_ppicture_deg; 
 XImage *x_ppicture_rest;
 
 char nomfen_ppicture_img[NCHAR];
 char nomfen_ppicture_deg[NCHAR];
 char nomfen_ppicture_rest[NCHAR];


 //Initialisation
 //--------------
 method=METHOD;

 psf=PSF;
 nbiter=ITERATION_REST;
 size=SIZE;
 variance_noise=VARIANCE_NOISE;

 lambda=LAMBDA;
 Tau=TAU;
 
 seed=SEED;
 
 zoom=ZOOM;
 flag_save=SAVE;
 flag_quit=QUIT;
 flag=0;

 //Lis Arguments
 //-------------
 read_arguments(argc,argv);


 //Affiche options
 //---------------
 printf("\n\n Info: ");
 printf("\n -----");
 printf("\n Ouverture de l'image : %s",Name_img);
 printf("\n -----");
 printf("\n Psf type ..............[%d]",psf);
 printf("\n Size of the Psf........[%d]",size);
 printf("\n Noise Variance.........[%.4f]",variance_noise); 
 printf("\n -----");
 printf("\n Restoration Method ....[%d]",method);
 printf("\n -----");
 //printf("\n Nb Iteration...........[%d]",nbiter);
 //printf("\n Lambda.................[%.3f]",lambda);
 //printf("\n Tau....................[%.2f]",Tau);
 //printf("\n -----");
 //printf("\n Value Seed.............[%d]",seed); 
 printf("\n -----");
 printf("\n Zoom...................[%d]",zoom); 
 if (flag_save)  printf("\n -> With Save");
 if (flag_quit)  printf("\n -> With Quit\n\n");

//-------------------------------
//---- Chargement des Images -------------------------------
//-------------------------------

 //Recuperation Images & Info
 //--------------------------
  int length,width;
  GetLengthWidth(Name_img,&length,&width);
  printf("\n Taille de l'image: [%d - %d]",length,width);
  fflush(stdout);
 
  int** datai=imatrix_allocate_2d(length,width);

  load_image_pgm(Name_img,datai,length,width);

  //------------------------------------------
  //-------- Algorithme INIT_RESTOR ----------
  //------------------------------------------
  Restor rest(datai,length,width);
 
  rest.option(nbiter,flag_save,lambda,Name_img,Tau);
  rest.restoration(method,psf,size,variance_noise);

  if (flag_save)
      { printf("\n\n Sauvegarde");
        SaveImagePgm("","RESTORED",rest.mat_rest,length,width); }

//---------------------------------------------------------------
//---------------- visu sous XWINDOW ----------------------------
//---------------------------------------------------------------
 if (!flag_quit)
 {
 //ouverture session graphique
 if (open_display()<0) printf(" Impossible d'ouvrir une session graphique");

 //commentaire dans les photos affichees
 sprintf(nomfen_ppicture_img,"Image : %s",Name_img);
 sprintf(nomfen_ppicture_deg,"Image Degrad : %s",Name_imgdeg);
 sprintf(nomfen_ppicture_rest,"%s",NAME_REST);
 
 //creation des fenetres
 win_ppicture_img=fabrique_window(nomfen_ppicture_img,10,10,width,length,zoom);
 win_ppicture_deg=fabrique_window(nomfen_ppicture_deg,10,10,width,length,zoom);
 win_ppicture_rest=fabrique_window(nomfen_ppicture_rest,10,10,width,length,zoom);
 
 //creation des Ximages
 x_ppicture_img=cree_Ximage4(datai,zoom,length,width);
 x_ppicture_deg=cree_Ximage2(rest.mat_deg,zoom,length,width);
 x_ppicture_rest=cree_Ximage2(rest.mat_rest,zoom,length,width);

 printf("\n\n Pour quitter,appuyer sur la barre d'espace");
 fflush(stdout);

 //boucle d'evenements
  for(;;)
     {
      XNextEvent(display,&ev);
       switch(ev.type)
        {
         case Expose:
         XPutImage(display,win_ppicture_img,gc,x_ppicture_img,0,0,0,0,
                   x_ppicture_img->width,x_ppicture_img->height);
         XPutImage(display,win_ppicture_deg,gc,x_ppicture_deg,0,0,0,0,
                   x_ppicture_deg->width,x_ppicture_deg->height);
         XPutImage(display,win_ppicture_rest,gc,x_ppicture_rest,0,0,0,0,
                   x_ppicture_rest->width,x_ppicture_rest->height);
         break;

         case KeyPress:
         XDestroyImage(x_ppicture_img);
         XDestroyImage(x_ppicture_deg);
         XDestroyImage(x_ppicture_rest);

         XFreeGC(display,gc);
         XCloseDisplay(display);
         flag=1;
         break;
         }
   if (flag==1) break;
   }
 } 
       
//--------------- fin de la session graphique --------------     
//----------------------------------------------------------

   //retour sans probleme
   //--------------------
   printf("\n C'est fini... \n");
   exit(-1);
   return 0;
 }
 


