//---------------------------------------------------
// module  : Restoration.h                         
// auteur  : Max Mignotte                            
// date    :                                              
// langage : C++                                      
// labo    :                                  
//---------------------------------------------------

#ifndef RESTORATION_H
#define RESTORATION_H

//------------------------------------------------
// FICHIERS INCLUS -------------------------------
//------------------------------------------------
#include "function.h"
#include "BuadesPrior.h"


//------------------------------------------------
// DEFINITIONS -----------------------------------
//------------------------------------------------
#define DEBUG 0
#define NCHAR 300

#define NAME_IMG_HIST      "ImgHist"
#define NAME_IMG_DEGRAD    "ImgDegrad"
#define NAME_IMG_RESTOR    "ImgRestor"

#define SEED            1
#define ITERATION_REST 10     
#define LAMBDA          1.0            
#define TAU             0.5

#define NB_GREY_LEVELS 256

#define MILLIONS 10000000
#define MILLE    1000

#define ARRETMAP 0.000005


//------------------------------------------------
// CLASSE RESTOR ---------------------------------
//------------------------------------------------
class  Restor
{
  //--Donnees--
  //-----------
  public:

  //Matrices
  float** mat_img;
  float** mat_deg;
  float** mat_rest;
  float** mat_pest;
  float** mat_est;
  float** mat_1est;
  float** mat_rint;
  float** mat_restbest;

  float** mat_psf;
  float** mat_tmp0; 
  float** mat_tmp1; 
  float** mat_tmp2; 
  float** mat_tmp3;
  float** mat_tmp4;
  float** mat_tmp5;
  float** mat_tmp6;

  float** matcg_q;

  int length;
  int width;
  int size_image;


  //Methode-ParRest.
  int   method;

  int   psf;
  int   size;
  int   halfsize;
  float variance_noise;

  int   nbiter;
  float lambda;
  float Tau;

  float SpeedGrad;
  int   seed;
 
  //Psf
  float*  SumCoefsPAR;
  float** CoefsGAUSS;
  float*  CoefsGAUSS2;

  //Mesures
  float bsnr_observed;
  float psnr_observed;
  float lambda_given;
  float isnr;
  float isnr_max;
  float isnr_stop;
  int   iter_isnr_max;  
  int   iter_isnr_stop;
  
  //Flag
  int  flag_save;
  int  flagR_silent;

  //Utile
  char Name_Rest[NCHAR];
  char Name_Sauve[NCHAR];
  char Name_Img[NCHAR];

  //--Methodes--
  //------------
  public:
  Restor(int**,int,int);
  ~Restor();

  //Choix Methode
  void restoration(int,int,int,float);   

  //Version Finale
  void UnsupNL_Rest2();

  //-------------------------------------------------
  //-------------------------------------------------
  //Options
  void option(int,int,float,char*,float); 

  //LandWeber
  void BoucleLandweber();   
  void GradientIterLandweber();         

  //Degradation
  void Degradation(float**,float**,float);                          

  //Mesure&Stat
  float ComputeLikelihoodMMSE(float**,float**,int);   
  float computeISNR(float**,float**,float**); 
  float computeMMSE(float**,float**,int); 
  float computePSNR(float**,float**,int);     
  float computeSNR(float**,float,int);  
  float computeSTAB(float**,float**,int);               
           
  //Gestion                
  void   ConvolutionSquareMatrixWithPsf(float**,float**,int); 
  void   SubstractSquareMatrix(float**,float**,float**,int);   
  void   AddSquareMatrix(float**,float**,float**,int); 
  void   AddSquareMatrix(float**,float**,float**,int,float); 
  void   AddSquareMatrix(float**,float,float**,float,float**,int);  
  void   MultSquareMatrix(float**,float**,float**,int);  
  void   FabsPowMatrix(float**,int,float);
  void   ConstraintMaxMin(float**,int);  
  float  MeanSqWindow(float**,int);
  float  DivNorm(float**,float**,int);

  //File
  void   saveImagePgm(char*,float**,int,int);   
  
  //-------------------------------------------------
  //--- Fonctions Amies  ----------------------------
  //-------------------------------------------------
  friend float*    fmatrix_allocate_1d(int);
  friend float**   fmatrix_allocate_2d(int,int);
  friend float***  fmatrix_allocate_3d(int,int,int);
  friend int*      imatrix_allocate_1d(int);
  friend int**     imatrix_allocate_2d(int,int);
  friend int***    imatrix_allocate_3d(int,int,int);
  friend void free_fmatrix_1d(float*);
  friend void free_fmatrix_2d(float**);
  friend void free_fmatrix_3d(float***,int);
  friend void free_imatrix_1d(int*);
  friend void free_imatrix_2d(int**);
  friend void free_imatrix_3d(int***,int);    
};

#endif 
