//------------------------------------------------------
// module  : function.cc
// auteur  : Mignotte Max
// date    :
// version : 1.0
// langage : C++
// labo    : DIRO
// note    :
//------------------------------------------------------
// 

//------------------------------------------------
// FICHIERS INCLUS -------------------------------
//------------------------------------------------
#include "function.h"


//--------------------------//
//-- Matrice de Flottant --//
//--------------------------//
//---------------------------------------------------------
//  alloue de la memoire pour une matrice 1d de float
//----------------------------------------------------------
float* fmatrix_allocate_1d(int hsize)
 {
  float* matrix;
  matrix=new float[hsize]; return matrix; }

//----------------------------------------------------------
//  alloue de la memoire pour une matrice 2d de float
//----------------------------------------------------------
float** fmatrix_allocate_2d(int vsize,int hsize)
 {
  float** matrix;
  float *imptr;

  matrix=new float*[vsize];
  imptr=new  float[(hsize)*(vsize)];
  for(int i=0;i<vsize;i++,imptr+=hsize) matrix[i]=imptr;
  return matrix;
 }

//----------------------------------------------------------
// alloue de la memoire pour une matrice 3d de float
//----------------------------------------------------------
float*** fmatrix_allocate_3d(int dsize,int vsize,int hsize)
 {
  float*** matrix;

  matrix=new float**[dsize];

  for(int i=0;i<dsize;i++)
    matrix[i]=fmatrix_allocate_2d(vsize,hsize);
  return matrix;
 }

//-----------------------//
//-- Matrice d'entiers --//
//-----------------------//
//---------------------------------------------------------
//  alloue de la memoire pour une matrice 1d de int
//----------------------------------------------------------
int* imatrix_allocate_1d(int hsize)
 {
  int* matrix;
  matrix=new int[hsize]; return matrix; }

//----------------------------------------------------------
//  alloue de la memoire pour une matrice 2d de int
//----------------------------------------------------------
int** imatrix_allocate_2d(int vsize,int hsize)
 {
  int** matrix;
  int*  imptr;

  matrix=new int*[vsize];
  imptr=new  int[(hsize)*(vsize)];
  for(int i=0;i<vsize;i++,imptr+=hsize) matrix[i]=imptr;
  return matrix;
 }

//----------------------------------------------------------
// alloue de la memoire pour une matrice 3d de int
//----------------------------------------------------------
int*** imatrix_allocate_3d(int dsize,int vsize,int hsize)
 {
  int*** matrix;

  matrix=new int**[dsize];

  for(int i=0;i<dsize;i++)
    matrix[i]=imatrix_allocate_2d(vsize,hsize);
  return matrix;
 }

//---------------------------//
//--- Liberation memoire ----//
//---------------------------//
//----------------------------------------------------------
// libere la memoire de la matrice 1d de float
//----------------------------------------------------------
void free_fmatrix_1d(float* pmat)
{ delete[] pmat; }

//----------------------------------------------------------
// libere la memoire de la matrice 2d de float
//----------------------------------------------------------
void free_fmatrix_2d(float** pmat)
{ delete[] (pmat[0]);
  delete[] pmat;}

//----------------------------------------------------------
// libere la memoire de la matrice 3d de float
//----------------------------------------------------------
void free_fmatrix_3d(float*** pmat,int dsize)
{
 for(int i=0;i<dsize;i++)
  {
   delete[] (pmat[i][0]);
   delete[] (pmat[i]);
   }
 delete[] (pmat);
}

//----------------------------------------------------------
// libere la memoire de la matrice 1d de int
//----------------------------------------------------------
void free_imatrix_1d(int* pmat)
{ delete[] pmat; }

//----------------------------------------------------------
// libere la memoire de la matrice 2d de int
//----------------------------------------------------------
void free_imatrix_2d(int** pmat)
{ delete[] (pmat[0]);
  delete[] pmat;}

//----------------------------------------------------------
// libere la memoire de la matrice 3d de int
//----------------------------------------------------------
void free_imatrix_3d(int*** pmat,int dsize)
{ for(int i=0;i<dsize;i++)
   {
    delete[] (pmat[i][0]);
    delete[] (pmat[i]);
    }
  delete[] (pmat); }


//--------------//
// BRUIT -------//
//--------------//
//----------------------------------------------------------
//  bruit gaussien  
//----------------------------------------------------------
float gaussian_noise(float var,float mean)
{
 float noise,theta;

 //generation du bruit
 noise=sqrt(-2*var*log(1.0-((float)rand()/RAND_MAX)));
 theta=(float)rand()*1.9175345E-4-PI;
 noise=noise*cos(theta);
 noise+=mean;
 if (noise>GREY_LEVEL) noise=GREY_LEVEL;
 if (noise<0) noise=0;
 return noise;
}

//----------------------------------------------------------
//  bruit gaussien sur la matrice 
//----------------------------------------------------------
void add_gaussian_noise(float** mat,int lgth,int wdth,float var)
{
 int i,j;

 //Boucle sur l'image
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
 if (var!=0.0) mat[i][j]=gaussian_noise(var,mat[i][j]);
}

//--------------------//
//-- LOAD/SAVE/FILE --//
//--------------------//
//----------------------------------------------------------
// Get Length and Width
//----------------------------------------------------------
void GetLengthWidth(char* path,int* length,int* width)
{
 unsigned char var;
 int   temp;
 char* tempc;
 char stringTmp1[100];
 char stringTmp2[100];
 int ta1,ta2;
 FILE *fic;

  //ouverture du fichier
  fic=fopen(path,"r");
  if (fic==NULL)
    { printf("\n- Grave erreur a l'ouverture de %s -\n",path);
      exit(-1); }

  //recuperation de l'entete
  tempc=fgets(stringTmp1,100,fic);
  for(;;) { temp=fread(&var,1,1,fic); if (var==35) tempc=fgets(stringTmp2,100,fic);
            else break; }
  fseek(fic,-1,SEEK_CUR);
  temp=fscanf(fic,"%d %d",&ta1,&ta2);
   
  //enregistrement
  (*length)=ta2; 
  (*width)=ta1;

  //fermeture du fichier
  fclose(fic);
}

//----------------------------------------------------------
// Chargement de l'image de nom <name> (en pgm)             
//----------------------------------------------------------
void load_image_pgm(char* path,int** img,int lgth,int wdth)
 {
  int i,j;
  int temp;
  char* tempc;
  unsigned char var;
  char header[100];
  char* ptr;
  int ta1,ta2,ta3;
  FILE *fic;

  //enregistrement du chemin;
   if (strstr(path,".pgm")==NULL) strcat(path,".pgm");

   //ouverture du fichier
  fic=fopen(path,"r");
  if (fic==NULL) 
   { printf("\n ->> Grave erreur a l'ouverture de %s -\n",path);
     exit(-1); }

  tempc=fgets(header,100,fic);
  if ( (header[0]!=80) ||    /* 'P' */
       (header[1]!=53) ) {   /* '5' */
       fprintf(stderr,"Image %s is not PGM.\n",path);
       exit(1); }

  tempc=fgets(header,100,fic);
  while(header[0]=='#') tempc=fgets(header,100,fic);
 
  ta1=strtol(header,&ptr,0);
  ta2=atoi(ptr);
  tempc=fgets(header,100,fic);
  ta3=strtol(header,&ptr,0);
   
  //chargement dans la matrice
  for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)  
    { temp=(int)fread(&var,1,1,fic);
      img[i][j]=(int)var; }

  //fermeture du fichier
  fclose(fic);
 }


//----------------------------------------------------------
// Sauvegarde de l'image de nom <name> au format pgm                        
//----------------------------------------------------------                
void SaveImagePgm(char* bruit,char* name,float** mat,int lgth,int wdth)
{
 int i,j;
 char buff[300];
 FILE* fic;

  //--extension--
  strcpy(buff,bruit);
  strcat(buff,name);
  strcat(buff,".pgm");

  //--ouverture fichier--
  fic=fopen(buff,"wb");
    if (fic==NULL) 
        { printf("Probleme dans la sauvegarde de %s",buff); 
          exit(-1); }
  printf("\n Sauvegarde de %s au format pgm\n",buff);

  //--sauvegarde de l'entete--
  fprintf(fic,"P5");
  fprintf(fic,"\n# IMG Module");
  fprintf(fic,"\n%d %d",wdth,lgth);
  fprintf(fic,"\n255\n");

  //--enregistrement--
  for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
	fprintf(fic,"%c",(char)mat[i][j]);
   
  //--fermeture fichier--
   fclose(fic); 
}


//-------------//
//-- GESTION --//
//-------------//
//----------------------------------------------------------
// copy_matrix
//----------------------------------------------------------
void copy_matrix(float** mat1,float** mat2,int lgth,int wdth)
{
 int i,j;

 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++)
   mat1[i][j]=mat2[i][j]; 
}

//----------------------------------------------------------
// CopySquareMatrix                 
//----------------------------------------------------------                
void CopySquareMatrix(float** mtin,float** mtout,int sz)
{
 int i,j;

 //Boucle
 for(i=0;i<sz;i++) for(j=0;j<sz;j++) mtout[i][j]=mtin[i][j];
}

//-------------//
//-- UTILE ----//
//-------------//
//----------------------------------------------------------
// Signe   
//----------------------------------------------------------  
float Signe(float val)
{
  float tmp=0.0; 

  //Condition
  if (val>0)  tmp=1.0;
  if (val<0)  tmp=-1.0;
  if (val==0) tmp=0.0;

  //Retour
  return tmp;
}

//----------------------------------------------------------*/
// ComputeNorm2                                              
//----------------------------------------------------------*/
float ComputeNorm2(float** mtin,float** mtout,int lgth,int wdth) 
{
 int i,j;
 float sum=0.0;

 //Boucle
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
   sum+=CARRE(mtin[i][j]-mtout[i][j]);

 sum/=(lgth*wdth);

 //Retour
 return sum; 
}

//----------------------------------------------------------*/
// ComputeNorm2                                              
//----------------------------------------------------------*/
float ComputeNorm2(float** mtin,int lgth,int wdth) 
{
 int i,j;
 float sum=0.0;

 //Boucle
 for(i=0;i<lgth;i++) for(j=0;j<wdth;j++) 
   sum+=CARRE(mtin[i][j]);

 sum/=(lgth*wdth);

 //Retour
 return sum; 
}






